/**
 *  \file RMF/external_decorators.h
 *  \brief Helper functions for manipulating RMF files.
 *
 *  Copyright 2007-2013 IMP Inventors. All rights reserved.
 *
 */

#ifndef RMF_EXTERNAL_DECORATORS_H
#define RMF_EXTERNAL_DECORATORS_H

#include <RMF/config.h>
#include <RMF/infrastructure_macros.h>
#include <RMF/NodeHandle.h>
#include <RMF/FileHandle.h>
#include <RMF/Decorator.h>
#include <RMF/Factory.h>
#include <RMF/constants.h>
#include <RMF/internal/utility.h>
#include <RMF/internal/paths.h>
#include <boost/array.hpp>
#include <boost/lexical_cast.hpp>

RMF_ENABLE_WARNINGS
namespace RMF {


  /** A reference to something in an external file. A relative path is stored.

       See also External and ExternalConstFactory.
    */
  class ExternalConst: public Decorator<NodeConstHandle> {
    friend class ExternalConstFactory;
    std::string get_name() const {return "External";}
  private:
    typedef Decorator<NodeConstHandle> P;
StringKey path_;
    ExternalConst(NodeConstHandle nh,
              StringKey path):
       P(nh),
path_(path) {
    }
  public:
  /** The absolute path to the external file. */
  String get_path() const {
    try {
      String relpath = get_node().get_value(path_);
      String filename = get_node().get_file().get_path();
      return internal::get_absolute_path(filename, relpath);
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "ExternalConst";
    }
  };

    typedef std::vector<ExternalConst> ExternalConsts;

  /** A reference to something in an external file. A relative path is stored.

       See also ExternalConst and ExternalFactory.
    */
  class External: public Decorator<NodeHandle> {
    friend class ExternalFactory;
    std::string get_name() const {return "External";}
  private:
    typedef Decorator<NodeHandle> P;
StringKey path_;
    External(NodeHandle nh,
              StringKey path):
       P(nh),
path_(path) {
    }
  public:
  /** The absolute path to the external file. */
  String get_path() const {
    try {
      String relpath = get_node().get_value(path_);
      String filename = get_node().get_file().get_path();
      return internal::get_absolute_path(filename, relpath);
    } RMF_DECORATOR_CATCH( );
  }
 /** The absolute path to the external file. */
  void set_path(String path) {
   try {
     String filename = get_node().get_file().get_path();
     String relpath = internal::get_relative_path(filename, path);
     get_node().set_value(path_, relpath);
   } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "External";
    }
  };

    typedef std::vector<External> Externals;

  /** Create decorators of type External.

       See also ExternalConst and ExternalFactory.
    */
  class ExternalConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
StringKey path_;

  public:
    ExternalConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("external")),
    path_(fh.get_key<StringTraits>(cat_, "path")) {
    }
    /** Get a ExternalConst for nh.*/
    ExternalConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  External");
      return ExternalConst(nh, path_);
    }
    /** Check whether nh has all the attributes required to be a
        ExternalConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(path_);
    }
    
  };

  typedef std::vector<ExternalConstFactory> ExternalConstFactories;

  /** Create decorators of type External.

       See also External and ExternalConstFactory.
    */
  class ExternalFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
StringKey path_;

  public:
    ExternalFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("external")),
    path_(fh.get_key<StringTraits>(cat_, "path")) {
    }
    /** Get a External for nh.*/
    External get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  External");
                                      
                                      
      return External(nh, path_);
    }
    /** Check whether nh has all the attributes required to be a
        External.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(path_);
    }
    
  };

  typedef std::vector<ExternalFactory> ExternalFactories;

} /* namespace RMF */
RMF_DISABLE_WARNINGS

#endif /* RMF_EXTERNAL_DECORATORS_H */
