/**
 *  \file RMF/sequence_decorators.h
 *  \brief Helper functions for manipulating RMF files.
 *
 *  Copyright 2007-2013 IMP Inventors. All rights reserved.
 *
 */

#ifndef RMF_SEQUENCE_DECORATORS_H
#define RMF_SEQUENCE_DECORATORS_H

#include <RMF/config.h>
#include <RMF/infrastructure_macros.h>
#include <RMF/NodeHandle.h>
#include <RMF/FileHandle.h>
#include <RMF/Decorator.h>
#include <RMF/Factory.h>
#include <RMF/constants.h>
#include <RMF/internal/utility.h>
#include <RMF/internal/paths.h>
#include <boost/array.hpp>
#include <boost/lexical_cast.hpp>

RMF_ENABLE_WARNINGS
namespace RMF {


  /** Information regarding a residue.

       See also Residue and ResidueConstFactory.
    */
  class ResidueConst: public Decorator<NodeConstHandle> {
    friend class ResidueConstFactory;
    std::string get_name() const {return "Residue";}
  private:
    typedef Decorator<NodeConstHandle> P;
boost::array<IntKey, 2> index_;
StringKey residue_type_;
    ResidueConst(NodeConstHandle nh,
              boost::array<IntKey, 2> index,
StringKey residue_type):
       P(nh),
index_(index),
residue_type_(residue_type) {
    }
  public:
  /** The index of the residue. */
  Int get_index() const {
    try {
      return get_node().get_value(index_[0]);
    } RMF_DECORATOR_CATCH( );
  }

  /** The three letter name for the residue. */
  String get_type() const {
    try {
      return P::get_value(residue_type_);
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "ResidueConst";
    }
  };

    typedef std::vector<ResidueConst> ResidueConsts;

  /** Information regarding a residue.

       See also ResidueConst and ResidueFactory.
    */
  class Residue: public Decorator<NodeHandle> {
    friend class ResidueFactory;
    std::string get_name() const {return "Residue";}
  private:
    typedef Decorator<NodeHandle> P;
boost::array<IntKey, 2> index_;
StringKey residue_type_;
    Residue(NodeHandle nh,
              boost::array<IntKey, 2> index,
StringKey residue_type):
       P(nh),
index_(index),
residue_type_(residue_type) {
    }
  public:
  /** The index of the residue. */
  Int get_index() const {
    try {
      return get_node().get_value(index_[0]);
    } RMF_DECORATOR_CATCH( );
  }
 /** The index of the residue. */
  void set_index(Int v) {
    try {
      get_node().set_value(index_[0], v);
      get_node().set_value(index_[1], v);
    } RMF_DECORATOR_CATCH( );
  }

  /** The three letter name for the residue. */
  String get_type() const {
    try {
      return P::get_value(residue_type_);
    } RMF_DECORATOR_CATCH( );
  }
  /** The three letter name for the residue. */
  void set_type(String v) {
    try {
      P::set_value(residue_type_, v);
    } RMF_DECORATOR_CATCH( );
  }


    static std::string get_decorator_type_name() {
         return "Residue";
    }
  };

    typedef std::vector<Residue> Residues;

  /** Create decorators of type Residue.

       See also ResidueConst and ResidueFactory.
    */
  class ResidueConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
boost::array<IntKey, 2> index_;
StringKey residue_type_;
  boost::array<IntKey, 2> get_index_keys(FileConstHandle fh) {
     boost::array<IntKey, 2> ret;
     ret[0] = fh.get_key<IntTraits>(cat_, "first residue index");
     ret[1] = fh.get_key<IntTraits>(cat_, "last residue index");
     return ret;
    }


  public:
    ResidueConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    index_(get_index_keys(fh)), residue_type_(fh.get_key<StringTraits>(cat_, "residue type")) {
    }
    /** Get a ResidueConst for nh.*/
    ResidueConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Residue");
      return ResidueConst(nh, index_,
residue_type_);
    }
    /** Check whether nh has all the attributes required to be a
        ResidueConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(index_[0]) && nh.get_has_value(index_[1]) && nh.get_value(index_[0]) == nh.get_value(index_[1])
    && nh.get_has_value(residue_type_);
    }
    

  };

  typedef std::vector<ResidueConstFactory> ResidueConstFactories;

  /** Create decorators of type Residue.

       See also Residue and ResidueConstFactory.
    */
  class ResidueFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
boost::array<IntKey, 2> index_;
StringKey residue_type_;
  boost::array<IntKey, 2> get_index_keys(FileHandle fh) {
     boost::array<IntKey, 2> ret;
     ret[0] = fh.get_key<IntTraits>(cat_, "first residue index");
     ret[1] = fh.get_key<IntTraits>(cat_, "last residue index");
     return ret;
    }


  public:
    ResidueFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    index_(get_index_keys(fh)), residue_type_(fh.get_key<StringTraits>(cat_, "residue type")) {
    }
    /** Get a Residue for nh.*/
    Residue get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Residue");
                                      
                                      
      return Residue(nh, index_,
residue_type_);
    }
    /** Check whether nh has all the attributes required to be a
        Residue.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(index_[0]) && nh.get_has_value(index_[1]) && nh.get_value(index_[0]) == nh.get_value(index_[1])
    && nh.get_has_value(residue_type_);
    }
    

  };

  typedef std::vector<ResidueFactory> ResidueFactories;


  /** Information regarding a chain.

       See also Chain and ChainConstFactory.
    */
  class ChainConst: public Decorator<NodeConstHandle> {
    friend class ChainConstFactory;
    std::string get_name() const {return "Chain";}
  private:
    typedef Decorator<NodeConstHandle> P;
IndexKey chain_id_;
    ChainConst(NodeConstHandle nh,
              IndexKey chain_id):
       P(nh),
chain_id_(chain_id) {
    }
  public:
  /** The one letter id for the chain (A is 0 etc). */
  Index get_chain_id() const {
    try {
      return P::get_value(chain_id_);
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "ChainConst";
    }
  };

    typedef std::vector<ChainConst> ChainConsts;

  /** Information regarding a chain.

       See also ChainConst and ChainFactory.
    */
  class Chain: public Decorator<NodeHandle> {
    friend class ChainFactory;
    std::string get_name() const {return "Chain";}
  private:
    typedef Decorator<NodeHandle> P;
IndexKey chain_id_;
    Chain(NodeHandle nh,
              IndexKey chain_id):
       P(nh),
chain_id_(chain_id) {
    }
  public:
  /** The one letter id for the chain (A is 0 etc). */
  Index get_chain_id() const {
    try {
      return P::get_value(chain_id_);
    } RMF_DECORATOR_CATCH( );
  }
  /** The one letter id for the chain (A is 0 etc). */
  void set_chain_id(Index v) {
    try {
      P::set_value(chain_id_, v);
    } RMF_DECORATOR_CATCH( );
  }


    static std::string get_decorator_type_name() {
         return "Chain";
    }
  };

    typedef std::vector<Chain> Chains;

  /** Create decorators of type Chain.

       See also ChainConst and ChainFactory.
    */
  class ChainConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
IndexKey chain_id_;

  public:
    ChainConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    chain_id_(fh.get_key<IndexTraits>(cat_, "chain id")) {
    }
    /** Get a ChainConst for nh.*/
    ChainConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Chain");
      return ChainConst(nh, chain_id_);
    }
    /** Check whether nh has all the attributes required to be a
        ChainConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(chain_id_);
    }
    
  };

  typedef std::vector<ChainConstFactory> ChainConstFactories;

  /** Create decorators of type Chain.

       See also Chain and ChainConstFactory.
    */
  class ChainFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
IndexKey chain_id_;

  public:
    ChainFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    chain_id_(fh.get_key<IndexTraits>(cat_, "chain id")) {
    }
    /** Get a Chain for nh.*/
    Chain get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Chain");
                                      
                                      
      return Chain(nh, chain_id_);
    }
    /** Check whether nh has all the attributes required to be a
        Chain.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(chain_id_);
    }
    
  };

  typedef std::vector<ChainFactory> ChainFactories;


  /** Information regarding a fragment of a molecule.

       See also Fragment and FragmentConstFactory.
    */
  class FragmentConst: public Decorator<NodeConstHandle> {
    friend class FragmentConstFactory;
    std::string get_name() const {return "Fragment";}
  private:
    typedef Decorator<NodeConstHandle> P;
IndexesKey indexes_;
    FragmentConst(NodeConstHandle nh,
              IndexesKey indexes):
       P(nh),
indexes_(indexes) {
    }
  public:
  /** A list of indexes in the fragment */
  Indexes get_indexes() const {
    try {
      return P::get_value(indexes_);
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "FragmentConst";
    }
  };

    typedef std::vector<FragmentConst> FragmentConsts;

  /** Information regarding a fragment of a molecule.

       See also FragmentConst and FragmentFactory.
    */
  class Fragment: public Decorator<NodeHandle> {
    friend class FragmentFactory;
    std::string get_name() const {return "Fragment";}
  private:
    typedef Decorator<NodeHandle> P;
IndexesKey indexes_;
    Fragment(NodeHandle nh,
              IndexesKey indexes):
       P(nh),
indexes_(indexes) {
    }
  public:
  /** A list of indexes in the fragment */
  Indexes get_indexes() const {
    try {
      return P::get_value(indexes_);
    } RMF_DECORATOR_CATCH( );
  }
  /** A list of indexes in the fragment */
  void set_indexes(Indexes v) {
    try {
      P::set_value(indexes_, v);
    } RMF_DECORATOR_CATCH( );
  }


    static std::string get_decorator_type_name() {
         return "Fragment";
    }
  };

    typedef std::vector<Fragment> Fragments;

  /** Create decorators of type Fragment.

       See also FragmentConst and FragmentFactory.
    */
  class FragmentConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
IndexesKey indexes_;

  public:
    FragmentConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    indexes_(fh.get_key<IndexesTraits>(cat_, "indexes")) {
    }
    /** Get a FragmentConst for nh.*/
    FragmentConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Fragment");
      return FragmentConst(nh, indexes_);
    }
    /** Check whether nh has all the attributes required to be a
        FragmentConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(indexes_);
    }
    
  };

  typedef std::vector<FragmentConstFactory> FragmentConstFactories;

  /** Create decorators of type Fragment.

       See also Fragment and FragmentConstFactory.
    */
  class FragmentFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
IndexesKey indexes_;

  public:
    FragmentFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    indexes_(fh.get_key<IndexesTraits>(cat_, "indexes")) {
    }
    /** Get a Fragment for nh.*/
    Fragment get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Fragment");
                                      
                                      
      return Fragment(nh, indexes_);
    }
    /** Check whether nh has all the attributes required to be a
        Fragment.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(indexes_);
    }
    
  };

  typedef std::vector<FragmentFactory> FragmentFactories;


  /** Information regarding a domain of a molecule.

       See also Domain and DomainConstFactory.
    */
  class DomainConst: public Decorator<NodeConstHandle> {
    friend class DomainConstFactory;
    std::string get_name() const {return "Domain";}
  private:
    typedef Decorator<NodeConstHandle> P;
boost::array<IntKey, 2> indexes_;
    DomainConst(NodeConstHandle nh,
              boost::array<IntKey, 2> indexes):
       P(nh),
indexes_(indexes) {
    }
  public:
  /** The range for the residues, specified as [first_index...last_index]. */
  IntRange get_indexes() const {
    try {
      return std::make_pair(get_node().get_value(indexes_[0]), get_node().get_value(indexes_[1]));
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "DomainConst";
    }
  };

    typedef std::vector<DomainConst> DomainConsts;

  /** Information regarding a domain of a molecule.

       See also DomainConst and DomainFactory.
    */
  class Domain: public Decorator<NodeHandle> {
    friend class DomainFactory;
    std::string get_name() const {return "Domain";}
  private:
    typedef Decorator<NodeHandle> P;
boost::array<IntKey, 2> indexes_;
    Domain(NodeHandle nh,
              boost::array<IntKey, 2> indexes):
       P(nh),
indexes_(indexes) {
    }
  public:
  /** The range for the residues, specified as [first_index...last_index]. */
  IntRange get_indexes() const {
    try {
      return std::make_pair(get_node().get_value(indexes_[0]), get_node().get_value(indexes_[1]));
    } RMF_DECORATOR_CATCH( );
  }
 /** The range for the residues, specified as [first_index...last_index]. */
  void set_indexes(IntRange v) {
    try {
      get_node().set_value(indexes_[0], v.first);
      get_node().set_value(indexes_[1], v.second);
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "Domain";
    }
  };

    typedef std::vector<Domain> Domains;

  /** Create decorators of type Domain.

       See also DomainConst and DomainFactory.
    */
  class DomainConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
boost::array<IntKey, 2> indexes_;
  boost::array<IntKey, 2> get_indexes_keys(FileConstHandle fh) {
     boost::array<IntKey, 2> ret;
     ret[0] = fh.get_key<IntTraits>(cat_, "first residue index");
     ret[1] = fh.get_key<IntTraits>(cat_, "last residue index");
     return ret;
    }

  public:
    DomainConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    indexes_(get_indexes_keys(fh)) {
    }
    /** Get a DomainConst for nh.*/
    DomainConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Domain");
      return DomainConst(nh, indexes_);
    }
    /** Check whether nh has all the attributes required to be a
        DomainConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(indexes_[0]) && nh.get_has_value(indexes_[1]) && nh.get_value(indexes_[0]) < nh.get_value(indexes_[1]);
    }
    
  };

  typedef std::vector<DomainConstFactory> DomainConstFactories;

  /** Create decorators of type Domain.

       See also Domain and DomainConstFactory.
    */
  class DomainFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
boost::array<IntKey, 2> indexes_;
  boost::array<IntKey, 2> get_indexes_keys(FileHandle fh) {
     boost::array<IntKey, 2> ret;
     ret[0] = fh.get_key<IntTraits>(cat_, "first residue index");
     ret[1] = fh.get_key<IntTraits>(cat_, "last residue index");
     return ret;
    }

  public:
    DomainFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    indexes_(get_indexes_keys(fh)) {
    }
    /** Get a Domain for nh.*/
    Domain get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Domain");
                                      
                                      
      return Domain(nh, indexes_);
    }
    /** Check whether nh has all the attributes required to be a
        Domain.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(indexes_[0]) && nh.get_has_value(indexes_[1]) && nh.get_value(indexes_[0]) < nh.get_value(indexes_[1]);
    }
    
  };

  typedef std::vector<DomainFactory> DomainFactories;


  /** A numeric tag for keeping track of types of molecules.

       See also Typed and TypedConstFactory.
    */
  class TypedConst: public Decorator<NodeConstHandle> {
    friend class TypedConstFactory;
    std::string get_name() const {return "Typed";}
  private:
    typedef Decorator<NodeConstHandle> P;
StringKey type_name_;
    TypedConst(NodeConstHandle nh,
              StringKey type_name):
       P(nh),
type_name_(type_name) {
    }
  public:
  /** An arbitrary tag representing the type. */
  String get_type_name() const {
    try {
      return P::get_value(type_name_);
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "TypedConst";
    }
  };

    typedef std::vector<TypedConst> TypedConsts;

  /** A numeric tag for keeping track of types of molecules.

       See also TypedConst and TypedFactory.
    */
  class Typed: public Decorator<NodeHandle> {
    friend class TypedFactory;
    std::string get_name() const {return "Typed";}
  private:
    typedef Decorator<NodeHandle> P;
StringKey type_name_;
    Typed(NodeHandle nh,
              StringKey type_name):
       P(nh),
type_name_(type_name) {
    }
  public:
  /** An arbitrary tag representing the type. */
  String get_type_name() const {
    try {
      return P::get_value(type_name_);
    } RMF_DECORATOR_CATCH( );
  }
  /** An arbitrary tag representing the type. */
  void set_type_name(String v) {
    try {
      P::set_value(type_name_, v);
    } RMF_DECORATOR_CATCH( );
  }


    static std::string get_decorator_type_name() {
         return "Typed";
    }
  };

    typedef std::vector<Typed> Typeds;

  /** Create decorators of type Typed.

       See also TypedConst and TypedFactory.
    */
  class TypedConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
StringKey type_name_;

  public:
    TypedConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    type_name_(fh.get_key<StringTraits>(cat_, "type name")) {
    }
    /** Get a TypedConst for nh.*/
    TypedConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Typed");
      return TypedConst(nh, type_name_);
    }
    /** Check whether nh has all the attributes required to be a
        TypedConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(type_name_);
    }
    
  };

  typedef std::vector<TypedConstFactory> TypedConstFactories;

  /** Create decorators of type Typed.

       See also Typed and TypedConstFactory.
    */
  class TypedFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
StringKey type_name_;

  public:
    TypedFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    type_name_(fh.get_key<StringTraits>(cat_, "type name")) {
    }
    /** Get a Typed for nh.*/
    Typed get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Typed");
                                      
                                      
      return Typed(nh, type_name_);
    }
    /** Check whether nh has all the attributes required to be a
        Typed.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(type_name_);
    }
    
  };

  typedef std::vector<TypedFactory> TypedFactories;


  /** Information regarding a copy of a molecule.

       See also Copy and CopyConstFactory.
    */
  class CopyConst: public Decorator<NodeConstHandle> {
    friend class CopyConstFactory;
    std::string get_name() const {return "Copy";}
  private:
    typedef Decorator<NodeConstHandle> P;
IndexKey copy_index_;
    CopyConst(NodeConstHandle nh,
              IndexKey copy_index):
       P(nh),
copy_index_(copy_index) {
    }
  public:
  /** This is the copy_indexth copy of the original. */
  Index get_copy_index() const {
    try {
      return P::get_value(copy_index_);
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "CopyConst";
    }
  };

    typedef std::vector<CopyConst> CopyConsts;

  /** Information regarding a copy of a molecule.

       See also CopyConst and CopyFactory.
    */
  class Copy: public Decorator<NodeHandle> {
    friend class CopyFactory;
    std::string get_name() const {return "Copy";}
  private:
    typedef Decorator<NodeHandle> P;
IndexKey copy_index_;
    Copy(NodeHandle nh,
              IndexKey copy_index):
       P(nh),
copy_index_(copy_index) {
    }
  public:
  /** This is the copy_indexth copy of the original. */
  Index get_copy_index() const {
    try {
      return P::get_value(copy_index_);
    } RMF_DECORATOR_CATCH( );
  }
  /** This is the copy_indexth copy of the original. */
  void set_copy_index(Index v) {
    try {
      P::set_value(copy_index_, v);
    } RMF_DECORATOR_CATCH( );
  }


    static std::string get_decorator_type_name() {
         return "Copy";
    }
  };

    typedef std::vector<Copy> Copys;

  /** Create decorators of type Copy.

       See also CopyConst and CopyFactory.
    */
  class CopyConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
IndexKey copy_index_;

  public:
    CopyConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    copy_index_(fh.get_key<IndexTraits>(cat_, "copy index")) {
    }
    /** Get a CopyConst for nh.*/
    CopyConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Copy");
      return CopyConst(nh, copy_index_);
    }
    /** Check whether nh has all the attributes required to be a
        CopyConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(copy_index_);
    }
    
  };

  typedef std::vector<CopyConstFactory> CopyConstFactories;

  /** Create decorators of type Copy.

       See also Copy and CopyConstFactory.
    */
  class CopyFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
IndexKey copy_index_;

  public:
    CopyFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("sequence")),
    copy_index_(fh.get_key<IndexTraits>(cat_, "copy index")) {
    }
    /** Get a Copy for nh.*/
    Copy get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Copy");
                                      
                                      
      return Copy(nh, copy_index_);
    }
    /** Check whether nh has all the attributes required to be a
        Copy.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(copy_index_);
    }
    
  };

  typedef std::vector<CopyFactory> CopyFactories;

} /* namespace RMF */
RMF_DISABLE_WARNINGS

#endif /* RMF_SEQUENCE_DECORATORS_H */
