/**
 *  \file RMF/shape_decorators.h
 *  \brief Helper functions for manipulating RMF files.
 *
 *  Copyright 2007-2013 IMP Inventors. All rights reserved.
 *
 */

#ifndef RMF_SHAPE_DECORATORS_H
#define RMF_SHAPE_DECORATORS_H

#include <RMF/config.h>
#include <RMF/infrastructure_macros.h>
#include <RMF/NodeHandle.h>
#include <RMF/FileHandle.h>
#include <RMF/Decorator.h>
#include <RMF/Factory.h>
#include <RMF/constants.h>
#include <RMF/internal/utility.h>
#include <RMF/internal/paths.h>
#include <boost/array.hpp>
#include <boost/lexical_cast.hpp>

RMF_ENABLE_WARNINGS
namespace RMF {


  /** These particles have associated color information.

       See also Colored and ColoredConstFactory.
    */
  class ColoredConst: public Decorator<NodeConstHandle> {
    friend class ColoredConstFactory;
    std::string get_name() const {return "Colored";}
  private:
    typedef Decorator<NodeConstHandle> P;
FloatKeys rgb_color_;
    ColoredConst(NodeConstHandle nh,
              FloatKeys rgb_color):
       P(nh),
rgb_color_(rgb_color) {
    }
  public:
  /** The RGB color. Each component has a value in [0...1]. */
  Floats get_rgb_color() const {
    try {
      Floats ret(rgb_color_.size());
      for (unsigned int i = 0; i < rgb_color_.size(); ++i) {
        ret[i] = P::get_value(rgb_color_[i]);
      }
      return ret;
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "ColoredConst";
    }
  };

    typedef std::vector<ColoredConst> ColoredConsts;

  /** These particles have associated color information.

       See also ColoredConst and ColoredFactory.
    */
  class Colored: public Decorator<NodeHandle> {
    friend class ColoredFactory;
    std::string get_name() const {return "Colored";}
  private:
    typedef Decorator<NodeHandle> P;
FloatKeys rgb_color_;
    Colored(NodeHandle nh,
              FloatKeys rgb_color):
       P(nh),
rgb_color_(rgb_color) {
    }
  public:
  /** The RGB color. Each component has a value in [0...1]. */
  Floats get_rgb_color() const {
    try {
      Floats ret(rgb_color_.size());
      for (unsigned int i = 0; i < rgb_color_.size(); ++i) {
        ret[i] = P::get_value(rgb_color_[i]);
      }
      return ret;
    } RMF_DECORATOR_CATCH( );
  }
  /** The RGB color. Each component has a value in [0...1]. */
  void set_rgb_color(Floats v) {
    try {
      for (unsigned int i = 0; i< rgb_color_.size(); ++i) {
         P::set_value(rgb_color_[i], v[i]);
      }
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "Colored";
    }
  };

    typedef std::vector<Colored> Coloreds;

  /** Create decorators of type Colored.

       See also ColoredConst and ColoredFactory.
    */
  class ColoredConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
FloatKeys rgb_color_;

  FloatKeys get_rgb_color_keys(FileConstHandle fh) {
    FloatKeys ret;
    ret.push_back(fh.get_key<FloatTraits>(cat_, "rgb color red"));;ret.push_back(fh.get_key<FloatTraits>(cat_, "rgb color green"));;ret.push_back(fh.get_key<FloatTraits>(cat_, "rgb color blue"));;
    return ret;
  }
  bool get_has_rgb_color_attributes(NodeConstHandle nh) const {
    for (unsigned int i = 0; i< rgb_color_.size(); ++i) {
      if (!nh.get_has_value(rgb_color_[i])) return false;
    }
    return true;
  }

  public:
    ColoredConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("shape")),
    rgb_color_(get_rgb_color_keys(fh)) {
    }
    /** Get a ColoredConst for nh.*/
    ColoredConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL||nh.get_type() == RMF::ALIAS||nh.get_type() == RMF::FEATURE||nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Colored");
      return ColoredConst(nh, rgb_color_);
    }
    /** Check whether nh has all the attributes required to be a
        ColoredConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL||nh.get_type() == RMF::ALIAS||nh.get_type() == RMF::FEATURE||nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(rgb_color_[0]) && nh.get_has_value(rgb_color_[1]) && nh.get_has_value(rgb_color_[2]);
    }
    
  };

  typedef std::vector<ColoredConstFactory> ColoredConstFactories;

  /** Create decorators of type Colored.

       See also Colored and ColoredConstFactory.
    */
  class ColoredFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
FloatKeys rgb_color_;

  FloatKeys get_rgb_color_keys(FileHandle fh) {
    FloatKeys ret;
    ret.push_back(fh.get_key<FloatTraits>(cat_, "rgb color red"));;ret.push_back(fh.get_key<FloatTraits>(cat_, "rgb color green"));;ret.push_back(fh.get_key<FloatTraits>(cat_, "rgb color blue"));;
    return ret;
  }
  bool get_has_rgb_color_attributes(NodeConstHandle nh) const {
    for (unsigned int i = 0; i< rgb_color_.size(); ++i) {
      if (!nh.get_has_value(rgb_color_[i])) return false;
    }
    return true;
  }

  public:
    ColoredFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("shape")),
    rgb_color_(get_rgb_color_keys(fh)) {
    }
    /** Get a Colored for nh.*/
    Colored get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL||nh.get_type() == RMF::ALIAS||nh.get_type() == RMF::FEATURE||nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Colored");
                                      
                                      
      return Colored(nh, rgb_color_);
    }
    /** Check whether nh has all the attributes required to be a
        Colored.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL||nh.get_type() == RMF::ALIAS||nh.get_type() == RMF::FEATURE||nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(rgb_color_[0]) && nh.get_has_value(rgb_color_[1]) && nh.get_has_value(rgb_color_[2]);
    }
    
  };

  typedef std::vector<ColoredFactory> ColoredFactories;


  /** A geometric ball.

       See also Ball and BallConstFactory.
    */
  class BallConst: public Decorator<NodeConstHandle> {
    friend class BallConstFactory;
    std::string get_name() const {return "Ball";}
  private:
    typedef Decorator<NodeConstHandle> P;
FloatKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    BallConst(NodeConstHandle nh,
              FloatKeys coordinates,
FloatKey radius,
IndexKey type):
       P(nh),
coordinates_(coordinates),
radius_(radius),
type_(type) {
    }
  public:
  /** The coordinates in angstroms. */
  Floats get_coordinates() const {
    try {
      Floats ret(coordinates_.size());
      for (unsigned int i = 0; i < coordinates_.size(); ++i) {
        ret[i] = P::get_value(coordinates_[i]);
      }
      return ret;
    } RMF_DECORATOR_CATCH( );
  }

  /** The radius in angstroms. */
  Float get_radius() const {
    try {
      return P::get_value(radius_);
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "BallConst";
    }
  };

    typedef std::vector<BallConst> BallConsts;

  /** A geometric ball.

       See also BallConst and BallFactory.
    */
  class Ball: public Decorator<NodeHandle> {
    friend class BallFactory;
    std::string get_name() const {return "Ball";}
  private:
    typedef Decorator<NodeHandle> P;
FloatKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    Ball(NodeHandle nh,
              FloatKeys coordinates,
FloatKey radius,
IndexKey type):
       P(nh),
coordinates_(coordinates),
radius_(radius),
type_(type) {
    }
  public:
  /** The coordinates in angstroms. */
  Floats get_coordinates() const {
    try {
      Floats ret(coordinates_.size());
      for (unsigned int i = 0; i < coordinates_.size(); ++i) {
        ret[i] = P::get_value(coordinates_[i]);
      }
      return ret;
    } RMF_DECORATOR_CATCH( );
  }
  /** The coordinates in angstroms. */
  void set_coordinates(Floats v) {
    try {
      for (unsigned int i = 0; i< coordinates_.size(); ++i) {
         P::set_value(coordinates_[i], v[i]);
      }
    } RMF_DECORATOR_CATCH( );
  }

  /** The radius in angstroms. */
  Float get_radius() const {
    try {
      return P::get_value(radius_);
    } RMF_DECORATOR_CATCH( );
  }
  /** The radius in angstroms. */
  void set_radius(Float v) {
    try {
      P::set_value(radius_, v);
    } RMF_DECORATOR_CATCH( );
  }


    static std::string get_decorator_type_name() {
         return "Ball";
    }
  };

    typedef std::vector<Ball> Balls;

  /** Create decorators of type Ball.

       See also BallConst and BallFactory.
    */
  class BallConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
FloatKeys coordinates_;
FloatKey radius_;
IndexKey type_;

  FloatKeys get_coordinates_keys(FileConstHandle fh) {
    FloatKeys ret;
    ret.push_back(fh.get_key<FloatTraits>(cat_, "cartesian x"));;ret.push_back(fh.get_key<FloatTraits>(cat_, "cartesian y"));;ret.push_back(fh.get_key<FloatTraits>(cat_, "cartesian z"));;
    return ret;
  }
  bool get_has_coordinates_attributes(NodeConstHandle nh) const {
    for (unsigned int i = 0; i< coordinates_.size(); ++i) {
      if (!nh.get_has_value(coordinates_[i])) return false;
    }
    return true;
  }


  public:
    BallConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("shape")),
    coordinates_(get_coordinates_keys(fh)), radius_(fh.get_key<FloatTraits>(cat_, "radius")), type_(fh.get_key<IndexTraits>(cat_, "type")) {
    }
    /** Get a BallConst for nh.*/
    BallConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Ball");
      return BallConst(nh, coordinates_,
radius_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        BallConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0]) && nh.get_has_value(coordinates_[1]) && nh.get_has_value(coordinates_[2])
    && nh.get_has_value(radius_)
    && nh.get_has_value(type_)
    && nh.get_value(type_) == 0;
    }
    
  /** Get values from a list of NodeIDs all at the same time. This method
      is a bit experimental and may change.
    */
  FloatsList get_coordinates(FileConstHandle fh,
                       const NodeIDs &node_ids) const {
      FloatsList ret(node_ids.size());
      for (unsigned int i = 0; i< node_ids.size(); ++i) {
        NodeConstHandle nh = fh.get_node(node_ids[i]);
        ret[i].resize(coordinates_.size());
        for (unsigned int j = 0; j < coordinates_.size(); ++j) {
          ret[i][j] = nh.get_value(coordinates_[j]);
        }
      }
      return ret;
  }


  };

  typedef std::vector<BallConstFactory> BallConstFactories;

  /** Create decorators of type Ball.

       See also Ball and BallConstFactory.
    */
  class BallFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
FloatKeys coordinates_;
FloatKey radius_;
IndexKey type_;

  FloatKeys get_coordinates_keys(FileHandle fh) {
    FloatKeys ret;
    ret.push_back(fh.get_key<FloatTraits>(cat_, "cartesian x"));;ret.push_back(fh.get_key<FloatTraits>(cat_, "cartesian y"));;ret.push_back(fh.get_key<FloatTraits>(cat_, "cartesian z"));;
    return ret;
  }
  bool get_has_coordinates_attributes(NodeConstHandle nh) const {
    for (unsigned int i = 0; i< coordinates_.size(); ++i) {
      if (!nh.get_has_value(coordinates_[i])) return false;
    }
    return true;
  }


  public:
    BallFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("shape")),
    coordinates_(get_coordinates_keys(fh)), radius_(fh.get_key<FloatTraits>(cat_, "radius")), type_(fh.get_key<IndexTraits>(cat_, "type")) {
    }
    /** Get a Ball for nh.*/
    Ball get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Ball");
                                      nh.set_value(type_, 0);
                                      
      return Ball(nh, coordinates_,
radius_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        Ball.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0]) && nh.get_has_value(coordinates_[1]) && nh.get_has_value(coordinates_[2])
    && nh.get_has_value(radius_)
    && nh.get_has_value(type_)
    && nh.get_value(type_) == 0;
    }
    
  /** Get values from a list of NodeIDs all at the same time. This method
      is a bit experimental and may change.
    */
  FloatsList get_coordinates(FileConstHandle fh,
                       const NodeIDs &node_ids) const {
      FloatsList ret(node_ids.size());
      for (unsigned int i = 0; i< node_ids.size(); ++i) {
        NodeConstHandle nh = fh.get_node(node_ids[i]);
        ret[i].resize(coordinates_.size());
        for (unsigned int j = 0; j < coordinates_.size(); ++j) {
          ret[i][j] = nh.get_value(coordinates_[j]);
        }
      }
      return ret;
  }


  };

  typedef std::vector<BallFactory> BallFactories;


  /** A geometric cylinder.

       See also Cylinder and CylinderConstFactory.
    */
  class CylinderConst: public Decorator<NodeConstHandle> {
    friend class CylinderConstFactory;
    std::string get_name() const {return "Cylinder";}
  private:
    typedef Decorator<NodeConstHandle> P;
FloatsKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    CylinderConst(NodeConstHandle nh,
              FloatsKeys coordinates,
FloatKey radius,
IndexKey type):
       P(nh),
coordinates_(coordinates),
radius_(radius),
type_(type) {
    }
  public:
  /** Coordinates of the center in angstroms. */
  FloatsList get_coordinates() const {
    try {
      FloatsList ret(coordinates_.size());
      for (unsigned int i = 0; i < coordinates_.size(); ++i) {
        ret[i] = P::get_value(coordinates_[i]);
      }
      return ret;
    } RMF_DECORATOR_CATCH( );
  }

  /** The radius in angstroms. */
  Float get_radius() const {
    try {
      return P::get_value(radius_);
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "CylinderConst";
    }
  };

    typedef std::vector<CylinderConst> CylinderConsts;

  /** A geometric cylinder.

       See also CylinderConst and CylinderFactory.
    */
  class Cylinder: public Decorator<NodeHandle> {
    friend class CylinderFactory;
    std::string get_name() const {return "Cylinder";}
  private:
    typedef Decorator<NodeHandle> P;
FloatsKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    Cylinder(NodeHandle nh,
              FloatsKeys coordinates,
FloatKey radius,
IndexKey type):
       P(nh),
coordinates_(coordinates),
radius_(radius),
type_(type) {
    }
  public:
  /** Coordinates of the center in angstroms. */
  FloatsList get_coordinates() const {
    try {
      FloatsList ret(coordinates_.size());
      for (unsigned int i = 0; i < coordinates_.size(); ++i) {
        ret[i] = P::get_value(coordinates_[i]);
      }
      return ret;
    } RMF_DECORATOR_CATCH( );
  }
  /** Coordinates of the center in angstroms. */
  void set_coordinates(FloatsList v) {
    try {
      for (unsigned int i = 0; i< coordinates_.size(); ++i) {
         P::set_value(coordinates_[i], v[i]);
      }
    } RMF_DECORATOR_CATCH( );
  }

  /** The radius in angstroms. */
  Float get_radius() const {
    try {
      return P::get_value(radius_);
    } RMF_DECORATOR_CATCH( );
  }
  /** The radius in angstroms. */
  void set_radius(Float v) {
    try {
      P::set_value(radius_, v);
    } RMF_DECORATOR_CATCH( );
  }


    static std::string get_decorator_type_name() {
         return "Cylinder";
    }
  };

    typedef std::vector<Cylinder> Cylinders;

  /** Create decorators of type Cylinder.

       See also CylinderConst and CylinderFactory.
    */
  class CylinderConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
FloatsKeys coordinates_;
FloatKey radius_;
IndexKey type_;

  FloatsKeys get_coordinates_keys(FileConstHandle fh) {
    FloatsKeys ret;
    ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian xs"));;ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian ys"));;ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian zs"));;
    return ret;
  }
  bool get_has_coordinates_attributes(NodeConstHandle nh) const {
    for (unsigned int i = 0; i< coordinates_.size(); ++i) {
      if (!nh.get_has_value(coordinates_[i])) return false;
    }
    return true;
  }


  public:
    CylinderConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("shape")),
    coordinates_(get_coordinates_keys(fh)), radius_(fh.get_key<FloatTraits>(cat_, "radius")), type_(fh.get_key<IndexTraits>(cat_, "type")) {
    }
    /** Get a CylinderConst for nh.*/
    CylinderConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Cylinder");
      return CylinderConst(nh, coordinates_,
radius_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        CylinderConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0]) && nh.get_has_value(coordinates_[1]) && nh.get_has_value(coordinates_[2])
    && nh.get_has_value(radius_)
    && nh.get_has_value(type_)
    && nh.get_value(type_) == 1;
    }
    

  };

  typedef std::vector<CylinderConstFactory> CylinderConstFactories;

  /** Create decorators of type Cylinder.

       See also Cylinder and CylinderConstFactory.
    */
  class CylinderFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
FloatsKeys coordinates_;
FloatKey radius_;
IndexKey type_;

  FloatsKeys get_coordinates_keys(FileHandle fh) {
    FloatsKeys ret;
    ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian xs"));;ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian ys"));;ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian zs"));;
    return ret;
  }
  bool get_has_coordinates_attributes(NodeConstHandle nh) const {
    for (unsigned int i = 0; i< coordinates_.size(); ++i) {
      if (!nh.get_has_value(coordinates_[i])) return false;
    }
    return true;
  }


  public:
    CylinderFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("shape")),
    coordinates_(get_coordinates_keys(fh)), radius_(fh.get_key<FloatTraits>(cat_, "radius")), type_(fh.get_key<IndexTraits>(cat_, "type")) {
    }
    /** Get a Cylinder for nh.*/
    Cylinder get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Cylinder");
                                      nh.set_value(type_, 1);
                                      
      return Cylinder(nh, coordinates_,
radius_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        Cylinder.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0]) && nh.get_has_value(coordinates_[1]) && nh.get_has_value(coordinates_[2])
    && nh.get_has_value(radius_)
    && nh.get_has_value(type_)
    && nh.get_value(type_) == 1;
    }
    

  };

  typedef std::vector<CylinderFactory> CylinderFactories;


  /** A geometric line setgment.

       See also Segment and SegmentConstFactory.
    */
  class SegmentConst: public Decorator<NodeConstHandle> {
    friend class SegmentConstFactory;
    std::string get_name() const {return "Segment";}
  private:
    typedef Decorator<NodeConstHandle> P;
FloatsKeys coordinates_;
IndexKey type_;
    SegmentConst(NodeConstHandle nh,
              FloatsKeys coordinates,
IndexKey type):
       P(nh),
coordinates_(coordinates),
type_(type) {
    }
  public:
  /** Coordinates of the center in angstroms. */
  FloatsList get_coordinates() const {
    try {
      FloatsList ret(coordinates_.size());
      for (unsigned int i = 0; i < coordinates_.size(); ++i) {
        ret[i] = P::get_value(coordinates_[i]);
      }
      return ret;
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "SegmentConst";
    }
  };

    typedef std::vector<SegmentConst> SegmentConsts;

  /** A geometric line setgment.

       See also SegmentConst and SegmentFactory.
    */
  class Segment: public Decorator<NodeHandle> {
    friend class SegmentFactory;
    std::string get_name() const {return "Segment";}
  private:
    typedef Decorator<NodeHandle> P;
FloatsKeys coordinates_;
IndexKey type_;
    Segment(NodeHandle nh,
              FloatsKeys coordinates,
IndexKey type):
       P(nh),
coordinates_(coordinates),
type_(type) {
    }
  public:
  /** Coordinates of the center in angstroms. */
  FloatsList get_coordinates() const {
    try {
      FloatsList ret(coordinates_.size());
      for (unsigned int i = 0; i < coordinates_.size(); ++i) {
        ret[i] = P::get_value(coordinates_[i]);
      }
      return ret;
    } RMF_DECORATOR_CATCH( );
  }
  /** Coordinates of the center in angstroms. */
  void set_coordinates(FloatsList v) {
    try {
      for (unsigned int i = 0; i< coordinates_.size(); ++i) {
         P::set_value(coordinates_[i], v[i]);
      }
    } RMF_DECORATOR_CATCH( );
  }

    static std::string get_decorator_type_name() {
         return "Segment";
    }
  };

    typedef std::vector<Segment> Segments;

  /** Create decorators of type Segment.

       See also SegmentConst and SegmentFactory.
    */
  class SegmentConstFactory: public Factory<FileConstHandle> {
    typedef Factory<FileConstHandle> P;
    Category cat_;
FloatsKeys coordinates_;
IndexKey type_;

  FloatsKeys get_coordinates_keys(FileConstHandle fh) {
    FloatsKeys ret;
    ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian xs"));;ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian ys"));;ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian zs"));;
    return ret;
  }
  bool get_has_coordinates_attributes(NodeConstHandle nh) const {
    for (unsigned int i = 0; i< coordinates_.size(); ++i) {
      if (!nh.get_has_value(coordinates_[i])) return false;
    }
    return true;
  }

  public:
    SegmentConstFactory(FileConstHandle fh):
    P(),
    cat_(fh.get_category("shape")),
    coordinates_(get_coordinates_keys(fh)), type_(fh.get_key<IndexTraits>(cat_, "type")) {
    }
    /** Get a SegmentConst for nh.*/
    SegmentConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Segment");
      return SegmentConst(nh, coordinates_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        SegmentConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0]) && nh.get_has_value(coordinates_[1]) && nh.get_has_value(coordinates_[2])
    && nh.get_has_value(type_)
    && nh.get_value(type_) == 1;
    }
    
  };

  typedef std::vector<SegmentConstFactory> SegmentConstFactories;

  /** Create decorators of type Segment.

       See also Segment and SegmentConstFactory.
    */
  class SegmentFactory: public Factory<FileHandle> {
    typedef Factory<FileHandle> P;
    Category cat_;
FloatsKeys coordinates_;
IndexKey type_;

  FloatsKeys get_coordinates_keys(FileHandle fh) {
    FloatsKeys ret;
    ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian xs"));;ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian ys"));;ret.push_back(fh.get_key<FloatsTraits>(cat_, "cartesian zs"));;
    return ret;
  }
  bool get_has_coordinates_attributes(NodeConstHandle nh) const {
    for (unsigned int i = 0; i< coordinates_.size(); ++i) {
      if (!nh.get_has_value(coordinates_[i])) return false;
    }
    return true;
  }

  public:
    SegmentFactory(FileHandle fh):
    P(),
    cat_(fh.get_category("shape")),
    coordinates_(get_coordinates_keys(fh)), type_(fh.get_key<IndexTraits>(cat_, "type")) {
    }
    /** Get a Segment for nh.*/
    Segment get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Segment");
                                      nh.set_value(type_, 1);
                                      
      return Segment(nh, coordinates_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        Segment.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0]) && nh.get_has_value(coordinates_[1]) && nh.get_has_value(coordinates_[2])
    && nh.get_has_value(type_)
    && nh.get_value(type_) == 1;
    }
    
  };

  typedef std::vector<SegmentFactory> SegmentFactories;

} /* namespace RMF */
RMF_DISABLE_WARNINGS

#endif /* RMF_SHAPE_DECORATORS_H */
